module(..., package.seeall)
local s = require'scheme'
local iface_mod = require'nsgifaceconf'
local conflib = require'nsgconflib'

local iptables = "/usr/local/sbin/iptables"

local chainNameVal = function (val)
                        if type(val) == 'string' and val:match("^[%l][%w%p]*$") then
                           return val
                        else
                           return nil, "word with alpha lower-case first item"
                        end
                     end

local withInverted = function (t)
   out = s.deepcopy(t)
   for _,v in pairs(t) do
      out[#out +1] = "! "..tostring(v)
   end
   return out
end

function genCheckValOrInverse(checkFunc)
   return function(val)
             if type(checkFunc(val)) == 'nil' and type(val)=='string' then
                local nval = val:gsub('^[%s]*![%s]*','')
                if type(checkFunc(nval)) == 'nil' then
                   local _ , res = checkFunc(nval)
                   res = res.." Sign '!' at start may be used for inversion."
                else
                   return "! "..checkFunc(nval)
                end
             else
                return checkFunc(val)
             end
          end
end

local function getProtocols()
  local protocols = conflib.getFile"/etc/protocols"
  protocols = '\n'..(protocols or '')
  local res = {'all'}
  for proto in protocols:gmatch"\n%s*([^%s#]+)%s+%d+%s+" do
    res[#res+1] = proto
    res[#res+1] = '!'..proto
  end
  return res
end

local function getIcmpType()
  local help = s.readpopen(iptables.." -p icmp -h")
  local res = {}
  for it in help:match"Valid ICMP Types:(.*)$":gmatch"[%w%-]+" do
    res[#res+1] = it
    res[#res+1] = '!'..it
  end
  return res
end

local criterionPriority = {
  "protocol",
  "protocol-num",
  "mac-source",
  "source",
  "destination",
  "in-interface",
  "out-interface",
  "source-port",
  "destination-port",
  "icmp-type",
   "explicit-matches",
  "extra-options",
  "target",
  "to-source",
  "to-destination",

   "set-class",
   "new",
   "hashmode",
   "clustermac",
   "total-nodes",
   "local-node",
   "hash-init",
   "connmark-params",
   "random",
   "set-dscp",
   "set-dscp-class",
   "ecn-tcp-remove",
   "log-level",
   "log-prefix",
   "log-tcp-sequence",
   "log-tcp-options",
   "log-ip-options",
   "log-uid",
   "set-mark",
   "set-xmark",
   "to-ports",
   "to",
   "rateest-name",
   "rateest-interval",
   "rateest-ewmalog",
   "reject-with",
   "set-mss",
   "clamp-mss-to-pmtu",
   "strip-options",
   "set-tos",
   "ttl-set",
   "ttl-dec",
   "ttl-inc",
   "ulog-nlgroup",
   "ulog-prefix",
   "ulog-cprange",
   "ulog-qthreshold",

  "extra-target",
}
local criterionPatterns = {
  protocol = "-p %s",
  ["protocol-num"] = "-p %d",
  ["mac-source"] = "--mac-source %s",
  source = "-s %s",
  destination = "-d %s",
  ["in-interface"] = "-i %s",
  ["out-interface"] = "-o %s",
  ["source-port"] = "--sport %s",
  ["destination-port"] = "--dport %s",
  ["icmp-type"] = "--icmp-type %s",
  ["extra-options"] = "%s",
  target = "-j %s",
  ["to-source"] = "--to-source %s",
  ["to-destination"] = "--to-destination %s",

  ["set-class"] = "--set-class %s",
  ["new"] = "--new",
  ["hashmode"] = "--hashmode %s",
  ["clustermac"] = "--clustermac %s",
  ["total-nodes"] = "--total-nodes %d",
  ["local-node"] = "--local-node %d",
  ["hash-init"] = "--hash-init %d",
  ["connmark-params"] = "%s",
  ["random"] = "--random",
  ["set-dscp"] = "--set-dscp %d",
  ["set-dscp-class"] = "--set-dscp-class %s" ,
  ["ecn-tcp-remove"] = "--ecn-tcp-remove",
  ["log-level"] = "--log-level %d",
  ["log-prefix"] = "--log-prefix",
  ["log-tcp-sequence"] = "--log-tcp-sequence",
  ["log-tcp-options"] = "--log-tcp-options",
  ["log-ip-options"] = "--log-ip-options",
  ["log-uid"] = "--log-uid",
  ["set-mark"] = "--set-mark %s",
  ["set-xmark"] = "--set-xmark %s",
  ["to-ports"] = "--to-ports %s",
  ["to"] = "--to %s",
  ["rateest-name"] = "--rateest-name %s",
  ["rateest-interval"] = "--rateest-interval %s",
  ["rateest-ewmalog"] = "--rateest-ewmalog %d",
  ["reject-with"] = "--reject-with %s",
  ["set-mss"] = "--set-mss %d",
  ["clamp-mss-to-pmtu"] = "--clamp-mss-to-pmtu",
  ["strip-options"] = "--strip-options %s",
  ["set-tos"] = "--set-tos %s",
  ["ttl-set"] = "--ttl-set %d",
  ["ttl-dec"] = "--ttl-dec %d",
  ["ttl-inc"] = "--ttl-inc %d",
  ["ulog-nlgroup"] = "--ulog-nlgroup %d",
  ["ulog-prefix"] = "--ulog-prefix %s",
  ["ulog-cprange"] = "--ulog-cprange %d",
  ["ulog-qthreshold"] = "--ulog-qthreshold %d",

  ["extra-target"] = "%s",
}

local explicitModulesParams = {
   addrtype = { params = "%s"},
   ah = { params = "%s"},
   comment = { comment = "--comment '%s'"},
   connbytes = { params = "%s"},
   connlimit = { params = "%s"},
   connmark = { params = "%s"},
   conntrack = { params = "%s"},
   dscp = { params = "%s"},
   ecn = { params = "%s"},
   esp = { params = "%s"},
   hashlimit = { params = "%s"},
   helper = { params = "%s"},
   iprange = { params = "%s"},
   length = { params = "%s"},
   limit = { params = "%s"},
   mac = { ["mac-source"] = "--mac-source %s"},
   mark = { mark = "--mark %s"},
   multiport = {
      ["source-ports"] = "--source-ports %s" ,
      ["destination-ports"] = "--destination-ports %s" ,
   },
   --owner
   physdev = { ["physdev-in"] = "--physdev-in %s",
               ["physdev-out"] =  "--physdev-out %s",
               ["physdev-is-in"] = "--physdev-is-in",
               ["physdev-is-out"] =  "--physdev-is-out",
               ["physdev-is-bridged"] = "--physdev-is-bridged"
            },
   pkttype = {["pkt-type"] = "--pkt-type %s"},
   policy ={
      dir = "--dir %s",
      pol = "--pol %s",
      strict = "--strict",
      reqid = "--reqid %s",
      spi = "--spi %s",
      proto = "--proto %s",
      mode = "--mode %s",
      ["tunnel-src"] = "--tunnel-src %s",
      ["tunnel-dst"] = "--tunnel-dst %s",
      next = "--next",
   },
   quota = {quota = "--quota %d"},
   rateest = { params = "%s"},
   realm = { params = "%s"},
   recent = { params = "%s"},
   sctp = { params = "%s"},
   set = { params = "%s"},
   state = { params = "%s"},
   statistic = { params = "%s"},
   string = { params = "%s"},
   tcpmss = { params = "%s"},
   time = { params = "%s"},
   tos = { params = "%s"},
   ttl = { params = "%s"},
   u32 = { params = "%s"},
   unclean = { params = ""},
}

local function splitVal(val)
  if (tostring(val)):sub(1,1) == '!' then
    return val:sub(2), '! '
  else
    return val, ""
  end
end

local function setFilterRulesAux(tbl, chain, conf)
   local err = {}
   local function xtest(com)
      com = " -t "..tbl.." "..com
      return os.execute(iptables..com.." 1>/dev/null  2>/dev/null")
   end
   local function x(com)
      com = " -t "..tbl.." "..com
      local res = s.readpopen(iptables..com.." 2>&1")
      if not res:match'^%s*$' then
         err[#err+1] = '\niptables '..com..'\n'..res:gsub("\nTry[^\n]*\n$", '')
      end
   end
   if chainNameVal(chain) and (xtest("-L "..chain) ~= 0) then
      --create userchain if not created before
      x("-N "..chain)
   end
   x("-F "..chain)
   if #err > 0 then
      return false, err[1]
   end

   for idx = 0, #conf do
      local rule = conf[idx]
      if rule then
         local com = {"-A "..chain}
         for _, crit in ipairs(criterionPriority) do
            local val = rule[crit]
            if crit == "in-interface" or crit == "out-interface" then
               -- may be interface aliases
               val = iface_mod.getIfaceName(val)
            elseif (crit == "target") and chainNameVal(val) and (xtest("-L "..val) ~= 0) then
               --precreate userchain if needed for jump to it
               x("-N "..chain)
               if #err > 0 then
                  return false, err[1]
               end
            elseif (crit == "explicit-matches") and type(val)=="table" then
               local exp={""}
               local good=false
               for mod,parms in pairs(val) do
                  if explicitModulesParams[mod] then
                     exp[#exp + 1] = "-m "..mod
                     if type(parms)=="table" then
                        for k,v in pairs(parms) do
                           local pat = explicitModulesParams[mod][k]
                           if v and pat then
                              local val, prefix = splitVal(v)
                              exp[#exp+1] = prefix..pat:format(val)
                              good = true
                           end
                        end
                     end
                  end
               end
               if good then com[#com+1] = table.concat(exp,' ') end
            end
            local pat = criterionPatterns[crit]
            if val and pat then
               local val, prefix = splitVal(val)
               com[#com+1] = prefix..pat:format(val)
            end
         end
         x(table.concat(com, ' '))
      end
   end
   if conf["default-target"] then
      x("-P "..chain..' '..conf["default-target"])
   end
   if #err > 0 then
      return false, table.concat(err)
   end
   -- try to free unneded resources
   os.execute([==[    {
                       cd /tmp;
                       TMPF=`mktemp`;
                       R=`uname -r`;
                       {
                          {
                             ls -t1 /lib/modules/${R}/kernel/net/bridge/netfilter/ ;
                             ls -t1 /lib/modules/${R}/kernel/net/netfilter/ ;
                             ls -t1 /lib/modules/${R}/kernel/net/ipv4/netfilter/ ;
                          } | grep -oe '^[^.]*';
                       } > $TMPF;
                       MOD='X'
                       i=0
                       while [ "${MOD}"'X' !=  'X' ]; do
                          MOD=`lsmod | tail -n +2 | grep -E '^[^[:space:]]*[[:space:]]+[^[:space:]]+[[:space:]]+0[[:space:]]+'  | grep -oe '^[^ ]*' | grep -F -f $TMPF ;`
                          if [ "${MOD}"'X' !=  'X' ]; then
                             rmmod ${MOD};
                          fi;
                             i=$(($i+1));
                             if [ "$i" -gt "7" ]; then
                                break;
                             fi
                       done;
                       rm $TMPF;
                    } 1>/dev/null 2>/dev/null; ]==])



   return true
end

--subitems added automatically
local filterAliasHookKey = {
  nat = {},
  filter = {},
  mangle = {},
}

--this is NOT scheme
--this is array with scheme corrections depends on chains
local iptableSchemeCorrections = {
   PREROUTING = {
      ["out-interface"] = s.nilType,
   },
   INPUT = {
      ["out-interface"] = s.nilType,
   },
   FORWARD = {},
   OUTPUT = {
      ["in-interface"] = s.nilType,
      ["mac-source"] = s.nilType,
   },
   POSTROUTING = {
      ["in-interface"] = s.nilType,
      ["mac-source"] = s.nilType,
   },
   USER = {}
}


--this is NOT scheme
--this is array with TARGET and HELP definitions for scheme generation
local iptableTargets = {
   nat = {
      _H = "nfFilterRule",
      PREROUTING = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DNAT',
         'DROP',
         'LOG',
         'MARK',
         'NETMAP',
         'NFLOG',
         'RATEEST',
         'REDIRECT',
         'RETURN',
         'TCPMSS',
         'ULOG',
         --'NFQUEUE',
      },
      POSTROUTING = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'LOG',
         'MARK',
         'MASQUERADE',
         'NETMAP',
         'NFLOG',
         'RATEEST',
         'RETURN',
         'SNAT',
         'TCPMSS',
         'ULOG',
         --'NFQUEUE',
      },
      OUTPUT = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DNAT',
         'DROP',
         'LOG',
         'MARK',
         'NETMAP',
         'NFLOG',
         'RATEEST',
         'REDIRECT',
         'RETURN',
         'TCPMSS',
         'ULOG',
         --'NFQUEUE',
      },
      USER = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DNAT',
         'DROP',
         'LOG',
         'MARK',
         'MASQUERADE',
         'NETMAP',
         'NFLOG',
         'RATEEST',
         'REDIRECT',
         'RETURN',
         'SNAT',
         'TCPMSS',
         'ULOG',
         --'NFQUEUE',
      },
   },
   filter = {
      _H = "nfFilterRule",
      FORWARD = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'REJECT',
         'RETURN',
         'TCPMSS',
         'ULOG',
         --'NFQUEUE',

      },

      INPUT = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'REJECT',
         'RETURN',
         'TCPMSS',
         'ULOG',
         --'NFQUEUE',
      },

      OUTPUT = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'REJECT',
         'RETURN',
         'TCPMSS',
         'ULOG',
         --'NFQUEUE',
      },
      USER = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'REJECT',
         'RETURN',
         'TCPMSS',
         'ULOG',
         --'NFQUEUE',
      },
   },
   mangle = {
      _H = "nfMangleRule",
      PREROUTING = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'DSCP',
         'ECN',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'RETURN',
         'TCPMSS',
         'TCPOPTSTRIP',
         'TOS',
         'TPROXY',
         'TTL',
         'ULOG',
         --'HL',  --only for IPv6
         --'NFQUEUE',
      },
      INPUT = {
         'ACCEPT',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'DSCP',
         'ECN',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'RETURN',
         'TCPMSS',
         'TCPOPTSTRIP',
         'TOS',
         'TTL',
         'ULOG',
         --'HL',  --only for IPv6
         --'NFQUEUE',
      },
      FORWARD = {
         'ACCEPT',
         'CLASSIFY',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'DSCP',
         'ECN',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'RETURN',
         'TCPMSS',
         'TCPOPTSTRIP',
         'TOS',
         'TTL',
         'ULOG',
         --'HL',  --only for IPv6
         --'NFQUEUE',
      },
      OUTPUT = {
         'ACCEPT',
         'CLASSIFY',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'DSCP',
         'ECN',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'RETURN',
         'TCPMSS',
         'TCPOPTSTRIP',
         'TTL',
         'ULOG',
         --'HL',  --only for IPv6
         --'NFQUEUE',
      },
      POSTROUTING = {
         'ACCEPT',
         'CLASSIFY',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'DSCP',
         'ECN',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'RETURN',
         'TCPMSS',
         'TCPOPTSTRIP',
         'TOS',
         'TTL',
         'ULOG',
         --'HL',  --only for IPv6
         --'NFQUEUE',
      },
      USER = {
         'ACCEPT',
         'CLASSIFY',
         'CLUSTERIP',
         'CONNMARK',
         'DROP',
         'DSCP',
         'ECN',
         'LOG',
         'MARK',
         'NFLOG',
         'RATEEST',
         'RETURN',
         'TCPMSS',
         'TCPOPTSTRIP',
         'TOS',
         'TPROXY',
         'TTL',
         'ULOG',
         --'HL',  --only for IPv6
         --'NFQUEUE',
      },
   },
}

local iptableTargetOptions = {
   CLASSIFY = {
      ["set-class"] = s.stringType{
         _H = "nfClassifySetclass",
         _P = -90,
      },
   },
   CLUSTERIP = {
      ["new"] = s.booleanType{
         _H = "nfClusteripNew",
         _P = -90,
      },
      ["hashmode"] = s.choiceType{
         _H = "nfClusteripHashmode",
         _A = {
            'sourceip',
            'sourceip-sourceport',
            'sourceip-sourceport-destport',
         },
         _P = -90,
      },
      ["clustermac"] = s.stringType{
         _V = s.macaddressVal,
         _H = "nfClusteripClustermac",
         _P = -90,
      },
      ["total-nodes"] = s.integerType{
         _H = "nfClusteripTotalnodes",
         _P = -90,
      },
      ["local-node"] = s.integerType{
         _H = "nfClusteripLocalnode",
         _P = -90,
      },
      ["hash-init"] = s.integerType{
         _H = "nfClusteripHashinit",
         _P = -90,
      },
   },
   CONNMARK = {
      ["connmark-params"] = s.stringType{
         _H = "nfConnmarkParams",
         _P = -90,
      },
   },
   DNAT = {
      ["to-destination"] = s.stringType{
         _H = "nfDnatTodestination",
         _P = -90,
      },
      ["random"] = s.booleanType{
         _H = "nfDnatRandom",
         _P = -90,
      },
   },
   -- 'DROP'
   DSCP   = {
      ["set-dscp"] = s.integerType{
         _H = "nfDscpSetdscp",
         _P = -90,
      },
      ["set-dscp-class"] = s.stringType{
         _H = "nfDscpSetdscpclass",
         _P = -90,
      },
   },
   ECN = {
      ["ecn-tcp-remove"] = s.booleanType{
         _H = "nfEcnEcntcpremove",
         _P = -90,
      },
   },
   --'HL',  --only for IPv6
   LOG = {
      ["log-level"] = s.integerType{
         _H = "nfLogLoglevel",
         _P = -90,
      },
      ["log-prefix"] = s.stringType{
         _H = "nfLogLogprefix",
         _P = -90,
      },
      ["log-tcp-sequence"] = s.booleanType{
         _H = "nfLogLogtcpsequence",
         _P = -90,
      },
      ["log-tcp-options"] = s.booleanType{
         _H = "nfLogLogtcpoptions",
         _P = -90,
      },
      ["log-ip-options"] = s.booleanType{
         _H = "nfLogLogipoptions",
         _P = -90,
      },
      ["log-uid"] = s.booleanType{
         _H = "nfLogLoguid",
         _P = -90,
      },
   },
   MARK = {
      ["set-mark"] = s.stringType{
         _H = "nfMarkSetmark",
         _P = -90,
      },
      ["set-xmark"] = s.stringType{
         _H = "nfMarkSetxmark",
         _P = -90,
      },
   },
   MASQUERADE = {
      ["to-ports"] = s.stringType{
         _H = "nfMasqueradeToports",
         _P = -90,
      },
      ["random"] = s.booleanType{
         _H = "nfMasqueradeRandom",
         _P = -90,
      },
   },
   NETMAP = {
      ["to"] = s.stringType{
         _H = "nfNetmapTo",
         _P = -90,
      },
   },
   --'NFLOG',
   --'NFQUEUE',
   RATEEST = {
      ["rateest-name"] = s.stringType{
         _H = "nfRateestName",
         _P = -90,
      },
      ["rateest-interval"] = s.stringType{
         _H = "nfRateestInterval",
         _P = -90,
      },
      ["rateest-ewmalog"] = s.integerType{
         _H = "nfRateestEwmalog",
         _P = -90,
      },
   },
   REDIRECT = {
      ["to-ports"] = s.stringType{
         _H = "nfRedirectToports",
         _P = -90,
      },
      ["random"] = s.booleanType{
         _H = "nfRedirectRandom",
         _P = -90,
      },
   },
   REJECT = {
      ["reject-with"] = s.choiceType{
         _H = "nfRejectwith",
         _A = {
            'icmp-net-unreachable',
            'icmp-host-unreachable',
            'icmp-port-unreachable',
            'icmp-proto-unreachable',
            'icmp-net-prohibited',
            'icmp-host-prohibited',
            'icmp-admin-prohibited',
         },
         _P = -90,
      },
   },
   --'RETURN',
   SNAT = {
      ["to-source"] = s.stringType{
         _H = "nfSnatTosource",
         _P = -90,
      },
      ["random"] = s.booleanType{
         _H = "nfSnatRandom",
         _P = -90,
      },
   },
   TCPMSS = {
      ["set-mss"] = s.integerType{
         _H = "nfTcpmssSetmss",
         _P = -90,
      },
      ["clamp-mss-to-pmtu"] = s.booleanType{
         _H = "nfTcpmssClamptopmtu",
         _P = -90,
      },
   },
   TCPOPTSTRIP = {
      ['strip-options'] = s.stringType{
         _H = "nfTcpoptstripOptions",
         _P = -90,
      },
   },
   TOS = {
      ['set-tos'] = s.stringType{
         _H = "nfTosSet",
         _P = -90,
      },
   },
   --'TPROXY',
   TTL = {
      ['ttl-set'] = s.stringType{
         _H = "nfTtlSet",
         _P = -90,
      },
      ['ttl-dec'] = s.stringType{
         _H = "nfTtlDec",
         _P = -90,
      },
      ['ttl-inc'] = s.stringType{
         _H = "nfTtlInc",
         _P = -90,
      },
   },
   ULOG = {
      ["ulog-nlgroup"] = s.integerType{
         _H = "nfUlogNlgroup",
         _P = -90,
         _A = {1,32}
      },
      ["ulog-prefix"] = s.stringType{
         _H = "nfUlogPrefix",
         _P = -90,
      },
      ["ulog-cprange"] = s.integerType{
         _H = "nfUlogCprange",
         _P = -90,
      },
      ["ulog-qthreshold"] = s.integerType{
         _H = "nfUlogQthreshold",
         _P = -90,
      },
   },
   EMPTY = {
   }
}

local iptableExplicitMatches = s.newScheme{
   addrtype = s.newScheme{
      _H = "nfCriterionAddrtype",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   ah = s.newScheme{
      _H = "nfCriterionAh",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   comment = s.newScheme{
      _H = "nfCriterionComment",
      ["comment"] =  s.stringType{_H = "nfCriterionComment"},},
   connbytes = s.newScheme{
      _H = "nfCriterionConnbytes",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   connlimit = s.newScheme{
      _H = "nfCriterionConnlimit",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   connmark = s.newScheme{
      _H = "nfCriterionConnmark",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   conntrack = s.newScheme{
      _H = "nfCriterionConntrack",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   dscp = s.newScheme{
      _H = "nfCriterionDscp",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   ecn = s.newScheme{
      _H = "nfCriterionEcn",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   esp = s.newScheme{
      _H = "nfCriterionEsp",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   hashlimit = s.newScheme{
      _H = "nfCriterionHashlimit",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   helper = s.newScheme{
      _H = "nfCriterionHelper",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   iprange = s.newScheme{
      _H = "nfCriterionIprange",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   length = s.newScheme{
      _H = "nfCriterionLength",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   limit = s.newScheme{
      _H = "nfCriterionLimit",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   mac = s.newScheme{
      _H = "nfCriterionMac",
      ["mac-source"] = s.stringType{
         _V = genCheckValOrInverse(s.macaddressVal),
         _H = "nfCriterionMacSource",
      },
   },
   mark = s.newScheme{
      _H = "nfCriterionMark",
      ["mark"] = s.stringType{ _H = "nfCriterionMark", },
   },
   multiport = s.newScheme{
      _H = "nfCriterionMultiport",
      ["source-ports"] = s.stringType{ _H = "nfCriterionSports", },
      ["destination-ports"] = s.stringType{ _H = "nfCriterionDports", },
   },
   --owner = s.newScheme{},
   physdev = s.newScheme{
      _H = "nfCriterionPhysdev",
      ["physdev-in"] = s.stringType{ _H = "nfCriterionPhysdevIn", },
      ["physdev-out"] = s.stringType{ _H = "nfCriterionPhysdevOut", },
      ["physdev-is-in"] = s.booleanType{ _H = "nfCriterionPhysdevIsin", },
      ["physdev-is-out"] = s.booleanType{ _H = "nfCriterionPhysdevIsout", },
      ["physdev-is-bridged"] = s.booleanType{ _H = "nfCriterionPhysdevIsbridged", },
   },
   pkttype = s.newScheme{
      _H = "nfCriterionPkttype",
      ["pkt-type"] = s.choiceType{
         _H = "nfCriterionPkttype",
         _A = withInverted({"unicast", "broadcast", "multicast"}),
      }
   },
   policy = s.newScheme{
      _H = "nfCriterionPolicy",
      ["dir"] = s.choiceType{
         _H = "nfCriterionPolicyDir",
         _A = {"in", "out"}
      },
      ["pol"] = s.choiceType{
         _H = "nfCriterionPolicyPol",
         _A = {"none", "ipsec"}
      },
      ["strict"] = s.booleanType{_H = "nfCriterionPolicyStrict",},
      ["reqid"] = s.stringType{_H = "nfCriterionPolicyReqid",},
      ["spi"] = s.stringType{_H = "nfCriterionPolicySpi",},
      ["proto"] = s.choiceType{
         _H = "nfCriterionPolicyProto",
         _A = withInverted{"ah", "esp", "ipcomp"},
      },
      ["mode"] = s.choiceType{
         _H = "nfCriterionPolicyMode",
         _A = withInverted{"tunnel", "transport"},
      },
      ["tunnel-src"] = s.stringType{_H = "nfCriterionPolicyTunnelsrc",},
      ["tunnel-dst"] = s.stringType{_H = "nfCriterionPolicyTunneldst",},
      ["next"] = s.booleanType{_H = "nfCriterionPolicyNext",},
   },
   ["quota"] = s.newScheme{
      _H = "nfCriterionQuota",
      ["quota"] = s.integerType{_H = "nfCriterionQuota",}
   },
   rateest = s.newScheme{
      _H = "nfCriterionRateest",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},
   },
   realm = s.newScheme{
      _H = "nfCriterionRealm",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   recent = s.newScheme{
      _H = "nfCriterionRecent",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   sctp = s.newScheme{
      _H = "nfCriterionSctp",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   set = s.newScheme{
      _H = "nfCriterionSet",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   state = s.newScheme{
      _H = "nfCriterionState",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   statistic = s.newScheme{
      _H = "nfCriterionStatistic",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   string = s.newScheme{
      _H = "nfCriterionString",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   tcpmss = s.newScheme{
      _H = "nfCriterionTcpmss",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   time = s.newScheme{
      _H = "nfCriterionTime",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   tos = s.newScheme{
      _H = "nfCriterionTos",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   ttl = s.newScheme{
      _H = "nfCriterionTtl",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
   u32 = s.newScheme{
      _H = "nfCriterionU32",
      ["params"] =  s.stringType{_H = "nfCriterionExplicitParams"},},
  --[[ unclean = s.booleanType{
      _H = "nfCriterionUnclean",
      },
--]]
}

local function setFilterRules(path, sch, conf, prevConf)
  if not s.confCmp(conf, prevConf) then
    s.confValidate(path, sch, conf, true) -- may be bad parms from previos conf
    conf = s.addDefault(path, sch, conf, true)
    local tbl = path[#path-1]
    local chain = path[#path]
    if type(filterAliasHookKey[tbl][chain]) == 'nil' then
       filterAliasHookKey[tbl][chain] = {}
    end
    iface_mod.setAliasHook(filterAliasHookKey[tbl][chain],
      conf[1] and
      function()
        return  setFilterRulesAux(tbl, chain, conf)
      end or nil
    )
    return setFilterRulesAux(tbl,chain, conf)
 end
 return true
end

local portRange = s.newScheme{
  _T = 'string',
  _V = function(val)
    if type(val) == 'number' and val >= 0 and val <= 0xffff and val%1==0 then
      return tostring(val)
    end
    if type(val) == 'string' then
      local negative, p1, p2 = val:match"^(!?)(%d%d?%d?%d?%d?)(:?%d?%d?%d?%d?%d?)$"
      if (
        p1 and tonumber(p1) <= 0xffff  and
        (p2=='' or p2:sub(1,1) == ':' and tonumber(p2:sub(2)) <= 0xffff)
      ) then
        return val
      end
    end
    return nil, "port range in format  port[:port]"
  end
}

local criterionScheme = s.newScheme{
   protocol = s.newScheme{
      _H = "nfCriterionProtocol",
      _T = "choice",
      _A = getProtocols(),
      _V = s.choiceVal,
      _P = 60,
      _DEP = 1,
   },
   ["protocol-num"] =
      function(_, path)
         return not C.getConfOrDefault(s.clonePath(path, -1, "protocol")) and
         s.integerType{
         _H = "nfCriterionProtocolNum",
         _A = {0, 0xffff},
         _P = 50,
      }
   end,
   ["in-interface"] = s.stringType{
      _H = "nfCriterionInInterface",
      _P = 40,
   },
   ["out-interface"] = s.stringType{
      _H = "nfCriterionOutInterface",
      _P = 30,
   },
   source = s.stringType{
      _H = "nfCriterionSource",
      _P = 20
   },
   destination = s.stringType{
      _H = "nfCriterionDestination",
      _P = 10,
   },
   ["source-port"] =
      function(_, path)
         local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
         return (proto == 'tcp' or proto == 'udp') and
         portRange{
         _H = "nfCriterionSport",
         _P = 8,
      }
   end,
   ["destination-port"] =
      function(_, path)
         local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
         return (proto == 'tcp' or proto == 'udp') and
         portRange{
         _H = "nfCriterionDport",
         _P = 7,
      }
   end,
   ["icmp-type"] =
      function(_, path)
         local proto = C.getConfOrDefault(s.clonePath(path, -1, "protocol"))
         return (proto == 'icmp') and
         s.newScheme{
         _H = "nfCriterionIcmpType",
         _T = "choice",
         _A = getIcmpType(),
         _V = s.choiceVal,
         _P = 7,
      }
   end,
   ["explicit-matches"] = iptableExplicitMatches{
      _H = "nfExplicitMatches",
      _P = 6,
   },
   ["extra-options"] = s.stringType{
      _H = "nfCriterionExtra",
      _P = 1,
   },
   ["extra-target"] = s.stringType{
      _H = "nfExtraTarget",
      _P = -100,
   },
}

local criterionParms  =
   function(_, path)
      local targets = iptableTargets[path[#path -2]][path[#path-1]]
         or iptableTargets[path[#path -2]]['USER'] or {}
      targets = s.deepcopy(targets)
      local conf = C.getConfByPath(s.clonePath(path, -2) or {}) or {}
      for k,_ in pairs(conf) do
         if (tostring(k)):match("^[%l]") and k~= "show" then
            targets[#targets + 1] = tostring(k)
         end
      end
      local sch1 = criterionScheme({
         _H = iptableTargets[path[#path -2 ]]._H,
         target = s.choiceType{
            _H = "nfTarget",
            _A = targets,
            _DEP = 1,
         },
      })

      local target = C.getConfByPath(s.clonePath(path, nil, "target")) or "EMPTY"
      local sch2 = s.newScheme(iptableTargetOptions[target] or {})
      local sch3 = s.newScheme(iptableSchemeCorrections[path[#path-1]] or iptableSchemeCorrections['USER'] or {})

      return sch1(sch2)(sch3)
   end

local chainScheme = s.newScheme{
  _Z = setFilterRules,
  _LIST = criterionParms,
  ["default-target"] = s.newScheme{
    _H = "nfDefaultTarget",
    _T = "choice",
    _A = {
      'ACCEPT',
      'DROP',
    },
    _V = s.choiceVal,
  },
}

local LO_ADDRESS = "127.0.0.1"
local loAddrScheme = s.stringType{
  _H = "nfCriterionSource",
  _V = function(val)
    return val == LO_ADDRESS and LO_ADDRESS or
    nil, LO_ADDRESS
  end,
  _D = LO_ADDRESS,
}
local onlyAcceptTargetScheme = s.newScheme{
  _H = "nfTarget",
  _T = "choice",
  _A = {
    'ACCEPT',
  },
  _V = s.choiceVal,
  _D = 'ACCEPT',
}
local filterInput0 = s.newScheme{
  _H = "nfFilterInputRule0",
  source = loAddrScheme,
  target = onlyAcceptTargetScheme,
}
local filterOutput0 = s.newScheme{
  _H = "nfFilterOutputRule0",
  destination = loAddrScheme,
  target = onlyAcceptTargetScheme,
}

filterScheme = s.newScheme{
  _H = "nfFilter",
  _Z = function() return true end,
  INPUT = chainScheme{
    _H = "nfFilterInput",
    [0] = filterInput0,
    _P = 30,
  },
  FORWARD = chainScheme{
    _H = "nfFilterForward",
    _P = 20,
  },
  OUTPUT =  chainScheme{
    _H = "nfFilterOutput",
    [0] = filterOutput0,
    _P = 10,
  },
  _NAME_TYPE = chainNameVal,
  _NAME_H = "nfUserChainPrompt",
  _NAME = s.newScheme{
     _H = "nfUserChain",
     _P = 5,
     _DEP = 1,
     _Z = setFilterRules,
     _LIST = criterionParms,
  },
  show = s.newScheme{
    _H = "nfFilterShow",
    _X = function()
      return('\n'..s.readpopen(iptables.."  -nvL -t filter 2>&1"))
    end,
    _R = 'user',
  }
}

--------------------------------------------------------------------------------
-- NAT

natScheme = s.newScheme{
  _H = "nfNat",
  _Z = function() return true end,

  PREROUTING =  chainScheme{
     _H = "nfNatPrerouting",
     _P = 30,
  },
  POSTROUTING =  chainScheme{
    _H = "nfNatPostrouting",
    _P = 20,
  },
  OUTPUT =  chainScheme{
    _H = "nfNatOutput",
    _P = 10,
  },
  _NAME_TYPE = chainNameVal,
  _NAME_H = "nfUserChainPrompt",
  _NAME = chainScheme{
     _H = "nfUserChain",
     _P = 5,
     _DEP = 1,
     _Z = setFilterRules,
     _LIST = criterionParms,
  },
  show = s.newScheme{
    _H = "nfNatShow",
    _X = function()
      return('\n'..s.readpopen(iptables.."  -nvL -t nat 2>&1"))
    end,
    _R = 'user',
  }
}
--------------------------------------------------------------------------------
-- MANGLE

mangleScheme = s.newScheme{
  _H = "nfMangle",
  _Z = function() return true end,
  PREROUTING = chainScheme{
    _H = "nfManglePrerouting",
    _P = 30,
  },
  INPUT =  chainScheme{
    _H = "nfMangleInput",
    _P = 25,
  },
  FORWARD =  chainScheme{
    _H = "nfMangleForward",
    _P = 20,
  },
  OUTPUT =  chainScheme{
    _H = "nfMangleOutput",
    _P = 15,
  },
  POSTROUTING =  chainScheme{
    _H = "nfManglePostrouting",
    _P = 10,
  },
  _NAME_TYPE = chainNameVal,
  _NAME_H = "nfUserChainPrompt",
  _NAME = s.newScheme{
     _H = "nfUserChain",
     _P = 5,
     _DEP = 1,
     _Z = setFilterRules,
     _LIST = criterionParms,
  },
  show = s.newScheme{
    _H = "nfMangleShow",
    _X = function()
      return('\n'..s.readpopen(iptables.."  -nvL -t mangle 2>&1"))
    end,
    _R = 'user',
  },
}

--------------------------------------------------------------------------------

showNetFilter = s.newScheme{
  _H = "nfShow",
  _X = function()
    return('\n'..s.readpopen(iptables.."-save 2>&1"))
  end
}

nfScheme = s.newScheme{
  _Z = function() return true end,
  filter = filterScheme,
  nat = natScheme,
  mangle = mangleScheme,
  show = showNetFilter,
}
